NAME_KEY_IDENTIFIER = "name"
METRIC_KEY_IDENTIFIER = "metric"
OUTPUT_NAME_KEY_IDENTIFIER = "output_name"
THRESHOLD_KEY_IDENTIFIER = "threshold"
CONDITION_KEY_IDENTIFIER = "condition"
DIMENSION_KEY_IDENTIFIER = "dimension"
DIMENSION_VALUE_KEY_IDENTIFIER = "dimension_value"

DETAILED_RESULTS_KEY_IDENTIFIER = "detailed_results"
EVALUATIONS_RESULTS_KEY_IDENTIFIER = "evaluations_results"
MESSAGE_KEY_IDENTIFIER = "message"
METRIC_VALUE_KEY_IDENTIFIER = "metric_value"
STATUS_KEY_IDENTIFIER = "status"
PASS_COUNT_KEY_IDENTIFIER = "pass_count"
FAIL_COUNT_KEY_IDENTIFIER = "fail_count"
ERROR_COUNT_KEY_IDENTIFIER = "error_count"



class ConfigError(Exception):
    """Error raised when any of the provided configs for evaluation is wrong"""

    pass


class ConditionEvaluationError(Exception):
    """Error raised when an issue is encountered while evaluating a quality standard."""

    pass


class InternalEvaluationError(Exception):
    """Error raised when errors are met within a group evaluation."""

    def __init__(self, evaluation_result, *args):
        self.evaluation_result = evaluation_result
        self.error_message = args[0]

    def __str__(self):
        return self.error_message



from typing import Tuple, Union

import numpy as np
import pandas as pd
from sklearn.linear_model import Lasso

from inspec_ai.analytics import telemetry_service


@telemetry_service.track_function_executed("get_predictive_features")
def get_predictive_features(
    x_train: Union[pd.DataFrame, np.ndarray],
    y_train: Union[pd.DataFrame, pd.Series, np.ndarray],
    alpha: float = 0.1,
) -> Tuple[Union[pd.DataFrame, np.ndarray], dict]:
    """Gets the predictive features of a dataset based on the coefficients returned by a Lasso model used as a baseline.

    Args:
        x_train: Numpy array or pandas Series. Training dataset.
        y_train: Numpy array or pandas Series. Predicted target values.
        alpha: The L2 regularization parameter for the Lasso model. If 0, then all the features will be used.

    Returns:
        The dataset with only the optimal features.
        The dictionary containing the coefficients used to establish the optimal features

    """
    _validate_inputs(x_train, y_train)

    model = Lasso(alpha=alpha)
    model.fit(x_train, y_train)
    lasso_coef = model.coef_

    optimal_features, coef_dict = _format_output(x_train, lasso_coef)

    return optimal_features, coef_dict


def _format_output(x_train, lasso_coef):

    if isinstance(x_train, pd.DataFrame):
        optimal_features = x_train.loc[:, lasso_coef != 0]
        coef_dict = {x_train.columns[index]: coef for index, coef in enumerate(lasso_coef)}

    else:
        optimal_features = x_train[:, (lasso_coef != 0)]
        coef_dict = dict(enumerate(lasso_coef))

    return optimal_features, coef_dict


def _validate_inputs(x_train: Union[pd.DataFrame, np.ndarray], y_train: Union[pd.DataFrame, np.ndarray]) -> None:
    if not isinstance(x_train, (pd.DataFrame, np.ndarray)):
        raise ValueError(f"x_train must be a pandas Dataframe or a numpy array, not {type(x_train)}")

    if not isinstance(y_train, (pd.DataFrame, pd.Series, np.ndarray)):
        raise ValueError(f"y_train must be a pandas Dataframe, a pandas Series or a numpy array, not {type(y_train)}")

    if len(x_train) != len(y_train):
        raise ValueError("The provided inputs must contain the same number of elements: x_train and y_train.")



from typing import List

from colorama import Fore
from colorama import init as colorama_init

from inspec_ai.utils.utils import _StrEnum


class Color(_StrEnum):
    GREEN = "green"
    RED = "red"


def enumerate_elements_in_a_sentence(str_items: List[str]) -> str:
    if len(str_items) == 1:
        return str_items[0]

    tmp = ", ".join(str_items[:-1])
    tmp += " and "
    tmp += str_items[-1]

    return tmp


def color_text_for_console(string: str, color: Color) -> str:
    """Adds the code to color the python console output when called with the print function."""
    colorama_init()  # To enable ANSI escape character sequences (for color) on Windows

    python_terminal_color_codes = {Color.GREEN: Fore.GREEN, Color.RED: Fore.RED}

    color_code = python_terminal_color_codes[color]

    return color_code + string + Fore.RESET



import json

from inspec_ai.utils.utils import _StrEnum


class Metric(_StrEnum):
    MAE = "mae"
    MASE = "mase"
    MAPE = "mape"
    AUC = "auc"
    F1 = "f1"
    PRECISION = "precision"
    RECALL = "recall"
    ACCURACY = "accuracy"


class Condition(_StrEnum):
    LOWER = "lt"
    GREATER = "gt"
    LOWER_EQUALS = "lte"
    GREATER_EQUALS = "gte"
    EQUALS = "eq"

    def convert_to_plain_english(self):
        condition_to_plain_english_dict = {
            Condition.LOWER.value: "lower than",
            Condition.GREATER.value: "greater than",
            Condition.LOWER_EQUALS.value: "lower or equal to",
            Condition.GREATER_EQUALS.value: "greater or equal to",
            Condition.EQUALS.value: "equal to",
        }
        return condition_to_plain_english_dict[self.value]


class Status(_StrEnum):
    PASSED = "passed"
    FAILED = "failed"
    ERROR = "error"


class EvaluationMode(_StrEnum):
    GLOBAL = "global"
    FOR_SINGLE_DIM_VALUE = "for single dim value"
    FOR_EACH_UNIQUE_DIM_VALUE = "for each unique dim value"
    FOR_EACH_DIM_VALUE_IN_LIST = "for each dim value in list"


class JsonEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, _StrEnum):
            return str(obj)
        else:
            return super(JsonEncoder, self).default(obj)



import json
import math
import warnings
from datetime import datetime
from pathlib import Path
from typing import List, Optional, Tuple, Union

import pandas as pd
from sklearn.metrics import accuracy_score, f1_score, mean_absolute_error, mean_absolute_percentage_error, precision_score, recall_score, roc_auc_score

from inspec_ai.analytics import telemetry_service
from inspec_ai.metrics.mase import mean_absolute_scaled_error
from inspec_ai.quality.constants import (
    CONDITION_KEY_IDENTIFIER,
    DETAILED_RESULTS_KEY_IDENTIFIER,
    DIMENSION_KEY_IDENTIFIER,
    DIMENSION_VALUE_KEY_IDENTIFIER,
    ERROR_COUNT_KEY_IDENTIFIER,
    EVALUATIONS_RESULTS_KEY_IDENTIFIER,
    FAIL_COUNT_KEY_IDENTIFIER,
    MESSAGE_KEY_IDENTIFIER,
    METRIC_KEY_IDENTIFIER,
    METRIC_VALUE_KEY_IDENTIFIER,
    NAME_KEY_IDENTIFIER,
    OUTPUT_NAME_KEY_IDENTIFIER,
    PASS_COUNT_KEY_IDENTIFIER,
    STATUS_KEY_IDENTIFIER,
    THRESHOLD_KEY_IDENTIFIER,
)
from inspec_ai.quality.enums import Condition, EvaluationMode, Metric, Status
from inspec_ai.quality.exceptions import ConditionEvaluationError, ConfigError, InternalEvaluationError
from inspec_ai.quality.quality_standard import QualityStandard
from inspec_ai.quality.utils import make_plain_english_inner_result, make_plain_english_result
from inspec_ai.quality.visualisation.visualisation import generate_main_html_report
from inspec_ai.utils.text_utils import Color, color_text_for_console, enumerate_elements_in_a_sentence

ACCEPTED_TYPES_SINGLE_VALUES = (int, float, str, bool)
ACCEPTED_TYPES_MULTIPLE_VALUES = (list,)
ACCEPTED_DIMENSION_VALUE_TYPES_ENUMERATION = enumerate_elements_in_a_sentence(
    [i.__name__ for i in ACCEPTED_TYPES_SINGLE_VALUES + ACCEPTED_TYPES_MULTIPLE_VALUES]
)


@telemetry_service.track_function_executed("evaluate_quality_standards")
def evaluate_quality_standards(
    dimensions_df: pd.DataFrame,
    y_true: Union[pd.DataFrame, pd.Series],
    y_pred: Union[pd.DataFrame, pd.Series],
    config: List[QualityStandard],
    print_results_in_console: bool = True,
    verbose: bool = False,
    save_html_report: bool = False,
    save_as_json: bool = False,
    output_folder: Union[Path, str] = None,
) -> dict:

    """Determines if the predictions satisfy quality standards.

        Args:
            dimensions_df: A dataframe containing the columns with values (e.g. product category or product) that will be
                used for aggregating metrics when running tests.
            y_true: Numpy array or pandas Series. Ground truth (correct) target values.
            y_pred: Numpy array or pandas Series. Predicted target values.
            config: A list of quality standards.
            print_results_in_console: Whether to print evaluation results in console.
            verbose: Whether to print the detailed results when available.
            save_html_report: Whether to produce a html report.
            save_as_json: Whether to save the evaluation results as a json file.
            output_folder: Path to the output directory for the json and html reports.

    Returns: A dict containing the detailed results of the tests.
    """

    evaluations_results = []
    output_folder = Path(output_folder) if output_folder else Path(".")

    for quality_standard in config:
        try:
            _check_dimension_config(dimensions_df, quality_standard)
            evaluation_result = _evaluate_metric(dimensions_df, y_true, y_pred, quality_standard)
            evaluations_results.append(evaluation_result)

        except Exception as e:
            message = f"{type(e).__name__}: {e}"
            warnings.warn(message)
            evaluation_result = dict()
            evaluation_result[METRIC_KEY_IDENTIFIER] = quality_standard.metric
            evaluation_result[OUTPUT_NAME_KEY_IDENTIFIER] = quality_standard.output_name
            evaluation_result[THRESHOLD_KEY_IDENTIFIER] = quality_standard.threshold
            evaluation_result[CONDITION_KEY_IDENTIFIER] = quality_standard.condition
            evaluation_result[DIMENSION_KEY_IDENTIFIER] = quality_standard.dimension
            evaluation_result[DIMENSION_VALUE_KEY_IDENTIFIER] = quality_standard.dimension_value
            evaluation_result[STATUS_KEY_IDENTIFIER] = Status.ERROR
            evaluation_result[MESSAGE_KEY_IDENTIFIER] = message

            if isinstance(e, InternalEvaluationError):
                evaluation_result[DETAILED_RESULTS_KEY_IDENTIFIER] = e.evaluation_result

            evaluations_results.append(evaluation_result)

    if save_as_json:
        output_path = output_folder / f"quality_standards_evaluation_{datetime.today().strftime('%Y_%m_%d_%H_%M_%S')}.json"
        with open(output_path, "w") as f:
            json.dump(evaluations_results, f)

    results = {EVALUATIONS_RESULTS_KEY_IDENTIFIER: evaluations_results}
    results.update(_count_statuses(evaluations_results))

    if print_results_in_console:
        _print_console_report(results, verbose)

    if save_html_report:
        generate_main_html_report(results, output_folder)

    return results


def _count_statuses(evaluations_results: List[dict]) -> dict:
    statuses_counts = dict()

    statuses_counts[PASS_COUNT_KEY_IDENTIFIER] = 0
    statuses_counts[FAIL_COUNT_KEY_IDENTIFIER] = 0
    statuses_counts[ERROR_COUNT_KEY_IDENTIFIER] = 0

    for result in evaluations_results:
        if result[STATUS_KEY_IDENTIFIER] == Status.ERROR:
            statuses_counts[ERROR_COUNT_KEY_IDENTIFIER] += 1

        if result[STATUS_KEY_IDENTIFIER] == Status.PASSED:
            statuses_counts[PASS_COUNT_KEY_IDENTIFIER] += 1

        elif result[STATUS_KEY_IDENTIFIER] == Status.FAILED:
            statuses_counts[FAIL_COUNT_KEY_IDENTIFIER] += 1

    return statuses_counts


def _check_dimension_config(dimensions_df: pd.DataFrame, quality_standard: QualityStandard):
    if quality_standard.dimension:
        if quality_standard.dimension not in dimensions_df.columns:
            raise ConfigError(
                f"The specified dimension '{quality_standard.dimension}' is missing from the columns in the dimensions dataframe, cannot evaluate"
            )

        if quality_standard.dimension_value:
            if isinstance(quality_standard.dimension_value, ACCEPTED_TYPES_SINGLE_VALUES):
                if quality_standard.dimension_value not in dimensions_df[quality_standard.dimension].unique():
                    raise ConfigError(
                        f"The dimension value '{quality_standard.dimension_value}' is not present in the specified dimension "
                        f"'{quality_standard.dimension}', cannot evaluate"
                    )

            elif isinstance(quality_standard.dimension_value, ACCEPTED_TYPES_MULTIPLE_VALUES):
                if len(set(quality_standard.dimension_value).intersection(dimensions_df[quality_standard.dimension].unique())) == 0:
                    raise ConfigError(
                        f"All the specified dimension values are missing in the specified dimension '{quality_standard.dimension}', cannot evaluate"
                    )

            else:
                raise ConfigError(
                    f"Detected the 'dimension' and 'dimension_value' properties, but 'dimension_value' is "
                    f"{type(quality_standard.dimension_value).__name__} "
                    f"which is not among accepted types: {ACCEPTED_DIMENSION_VALUE_TYPES_ENUMERATION}"
                )
    else:
        if quality_standard.dimension_value:
            raise ConfigError("Detected the 'dimension_value' property, but missing 'dimension' property in configuration, cannot evaluate")


def _evaluate_metric(
    dimensions_df: pd.DataFrame, y_true: Union[pd.DataFrame, pd.Series], y_pred: Union[pd.DataFrame, pd.Series], quality_standard: QualityStandard
) -> dict:
    eval_result = {
        METRIC_KEY_IDENTIFIER: quality_standard.metric,
        CONDITION_KEY_IDENTIFIER: quality_standard.condition,
        THRESHOLD_KEY_IDENTIFIER: quality_standard.threshold,
        NAME_KEY_IDENTIFIER: quality_standard.name,
    }

    if quality_standard.output_name and isinstance(y_pred, pd.DataFrame) and isinstance(y_true, pd.DataFrame):
        y_true = y_true[quality_standard.output_name]
        y_pred = y_pred[quality_standard.output_name]
        eval_result[OUTPUT_NAME_KEY_IDENTIFIER] = quality_standard.output_name

    if quality_standard.evaluation_mode in [EvaluationMode.FOR_EACH_DIM_VALUE_IN_LIST, EvaluationMode.FOR_EACH_UNIQUE_DIM_VALUE]:
        if quality_standard.evaluation_mode == EvaluationMode.FOR_EACH_UNIQUE_DIM_VALUE:
            dimension_value = dimensions_df[quality_standard.dimension].unique().tolist()
        else:
            dimension_value = quality_standard.dimension_value

        eval_result[DETAILED_RESULTS_KEY_IDENTIFIER], eval_result[STATUS_KEY_IDENTIFIER] = _measure_metric_and_evaluate_internal_multiple(
            dimensions_df,
            y_true,
            y_pred,
            quality_standard.metric,
            quality_standard.condition,
            quality_standard.threshold,
            quality_standard.dimension,
            dimension_value,
        )

    else:
        dimension_value = quality_standard.dimension_value

        eval_result[METRIC_VALUE_KEY_IDENTIFIER], eval_result[STATUS_KEY_IDENTIFIER] = _measure_metric_and_evaluate_internal_single(
            dimensions_df,
            y_true,
            y_pred,
            quality_standard.metric,
            quality_standard.condition,
            quality_standard.threshold,
            quality_standard.dimension,
            dimension_value,
        )

    if quality_standard.dimension:
        eval_result[DIMENSION_KEY_IDENTIFIER] = quality_standard.dimension
    if quality_standard.dimension_value:
        eval_result[DIMENSION_VALUE_KEY_IDENTIFIER] = quality_standard.dimension_value

    return eval_result


def _measure_metric(
    dimensions_df: pd.DataFrame,
    y_true: pd.Series,
    y_pred: pd.Series,
    metric: Metric,
    dimension: Optional[Union[str, int, float, bool]] = None,
    dimension_value: Optional[Union[list, str, int, float, bool]] = None,
) -> float:
    y_tmp = y_true
    predictions_tmp = y_pred

    if dimension is not None and dimension_value is not None:
        # if a dimension is specified, slice y to only evaluate the metric for the chosen dimension value
        y_tmp = y_true[dimensions_df[dimension] == dimension_value]
        predictions_tmp = y_pred[dimensions_df[dimension] == dimension_value]

    if metric == Metric.MAE.value:
        return mean_absolute_error(y_tmp, predictions_tmp)

    if metric == Metric.MASE.value:
        return mean_absolute_scaled_error(y_tmp, predictions_tmp)

    if metric == Metric.MAPE.value:
        return mean_absolute_percentage_error(y_tmp, predictions_tmp)

    if metric == Metric.AUC.value:
        return roc_auc_score(y_tmp, predictions_tmp)

    if metric == Metric.F1.value:
        return f1_score(y_tmp, predictions_tmp)

    if metric == Metric.PRECISION.value:
        return precision_score(y_tmp, predictions_tmp)

    if metric == Metric.RECALL.value:
        return recall_score(y_tmp, predictions_tmp)

    else:
        return accuracy_score(y_tmp, predictions_tmp)


def _evaluate_internal(metric_value: Union[float, int], condition: Condition, threshold: Union[float, int]) -> bool:

    # lt = lower than
    if condition == Condition.LOWER:
        return metric_value < threshold

    # gt = greater than
    elif condition == Condition.GREATER:
        return metric_value > threshold

    # lte = lower than or equals
    elif condition == Condition.LOWER_EQUALS:
        return metric_value <= threshold

    # gte = greater than or equals
    elif condition == Condition.GREATER_EQUALS:
        return metric_value >= threshold

    # default is eq = equals
    else:
        return metric_value == threshold


def _measure_metric_and_evaluate_internal_single(
    dimensions_df: pd.DataFrame,
    y_true: pd.Series,
    y_pred: pd.Series,
    metric: Metric,
    condition: Condition,
    threshold: Union[float, int],
    dimension: Optional[Union[str, int, float, bool]],
    dimension_value: Optional[Union[str, int, float, bool]],
) -> Tuple[float, str]:

    metric_value = _measure_metric(dimensions_df, y_true, y_pred, metric, dimension, dimension_value)
    evaluation_success = _evaluate_internal(metric_value, condition, threshold)

    if math.isnan(metric_value):
        raise ConditionEvaluationError(f"The {metric} was evaluated to NaN.")

    status = Status.PASSED if evaluation_success else Status.FAILED

    return metric_value, status


def _check_config_for_multiple_dimension_values(dimensions_df, dimension, dimension_value):
    if dimension_value not in dimensions_df[dimension].unique().tolist():
        raise ConfigError(f"the dimension value '{dimension_value}' is missing in the dimension '{dimension}'")


def _measure_metric_and_evaluate_internal_multiple(
    dimensions_df: pd.DataFrame,
    y_true: pd.Series,
    y_pred: pd.Series,
    metric: Metric,
    condition: Condition,
    threshold: Union[float, int],
    dimension: Optional[Union[str, int, float, bool]],
    dimension_value: List[Union[str, int, float, bool]],
) -> Tuple[List[dict], str]:

    detailed_results = []

    for i in dimension_value:

        try:
            _check_config_for_multiple_dimension_values(dimensions_df, dimension, i)
            metric_value = _measure_metric(dimensions_df, y_true, y_pred, metric, dimension, i)
            evaluation_success = _evaluate_internal(metric_value, condition, threshold)
            status = Status.ERROR if math.isnan(metric_value) else Status.PASSED if evaluation_success else Status.FAILED
            message = ""

        except Exception as e:
            metric_value = None
            status = Status.ERROR
            message = f"{type(e).__name__}: {e}"

        detailed_results.append(
            {DIMENSION_VALUE_KEY_IDENTIFIER: i, METRIC_VALUE_KEY_IDENTIFIER: metric_value, STATUS_KEY_IDENTIFIER: status, MESSAGE_KEY_IDENTIFIER: message}
        )

    statuses_counts = _count_statuses(detailed_results)

    if statuses_counts[FAIL_COUNT_KEY_IDENTIFIER] > 0:
        global_status = Status.FAILED
    elif statuses_counts[ERROR_COUNT_KEY_IDENTIFIER] > 0:
        raise InternalEvaluationError(detailed_results, "Error(s) occurred when evaluating the quality standard for at least one group of dimension.")
    else:
        global_status = Status.PASSED

    return detailed_results, global_status


def _format_status_for_console_report(status: str) -> str:
    if status == Status.ERROR or status == Status.FAILED:
        return color_text_for_console(f"[{status}]", Color.RED)

    if status == Status.PASSED:
        return color_text_for_console(f"[{status}]", Color.GREEN)

    return f"[{status}]"


def _print_console_report(results: dict, verbose: bool):

    n_evaluations = len(results[EVALUATIONS_RESULTS_KEY_IDENTIFIER])
    completion_pct = round(100 * results[PASS_COUNT_KEY_IDENTIFIER] / n_evaluations, 2)

    print("QUALITY STANDARDS EVALUATION")

    print(
        f"Total - Pass: {color_text_for_console(str(results[PASS_COUNT_KEY_IDENTIFIER]), Color.GREEN)} "
        f"({str(completion_pct)}%); "
        f"Fail: {color_text_for_console(str(results[FAIL_COUNT_KEY_IDENTIFIER]), Color.RED)}; "
        f"Error: {str(results[ERROR_COUNT_KEY_IDENTIFIER])}"
    )

    print("===============================================================")

    for result in results[EVALUATIONS_RESULTS_KEY_IDENTIFIER]:

        if result[STATUS_KEY_IDENTIFIER] == Status.ERROR and DETAILED_RESULTS_KEY_IDENTIFIER not in result.keys():
            print(f"{_format_status_for_console_report(result[STATUS_KEY_IDENTIFIER])} - {result[MESSAGE_KEY_IDENTIFIER]}")
            print("===============================================================")

        else:
            print(
                f"{_format_status_for_console_report(result[STATUS_KEY_IDENTIFIER])} - "
                f"{result.get(NAME_KEY_IDENTIFIER) if result.get(NAME_KEY_IDENTIFIER) else make_plain_english_result(result)}"
                f"{' (' + str((round(result[METRIC_VALUE_KEY_IDENTIFIER], 4))) + ')' if METRIC_VALUE_KEY_IDENTIFIER in result.keys() else ''}"
            )

            if verbose:
                # print detailed result for evaluations over multiple dimension_values
                if DETAILED_RESULTS_KEY_IDENTIFIER in result.keys():

                    statuses_counts = _count_statuses(result[DETAILED_RESULTS_KEY_IDENTIFIER])

                    print("\n---------  DETAILED RESULTS  ----------------------------------")
                    print(
                        f"Total - Pass: {color_text_for_console(str(statuses_counts[PASS_COUNT_KEY_IDENTIFIER]), Color.GREEN)} "
                        f"({round(statuses_counts[PASS_COUNT_KEY_IDENTIFIER] / len(result[DETAILED_RESULTS_KEY_IDENTIFIER]) * 100, 1)}%); "
                        f"Fail: {color_text_for_console(str(statuses_counts[FAIL_COUNT_KEY_IDENTIFIER]), Color.RED)}; "
                        f"Error: {str(statuses_counts[ERROR_COUNT_KEY_IDENTIFIER])}"
                    )

                    for inner_result in result[DETAILED_RESULTS_KEY_IDENTIFIER]:
                        print(f"{_format_status_for_console_report(inner_result[STATUS_KEY_IDENTIFIER])} - {make_plain_english_inner_result(inner_result)}")

            print("===============================================================")



from inspec_ai.quality.constants import (
    CONDITION_KEY_IDENTIFIER,
    DIMENSION_KEY_IDENTIFIER,
    DIMENSION_VALUE_KEY_IDENTIFIER,
    MESSAGE_KEY_IDENTIFIER,
    METRIC_KEY_IDENTIFIER,
    METRIC_VALUE_KEY_IDENTIFIER,
    OUTPUT_NAME_KEY_IDENTIFIER,
    THRESHOLD_KEY_IDENTIFIER,
)
from inspec_ai.utils.text_utils import enumerate_elements_in_a_sentence


def make_plain_english_result(evaluation_result: dict) -> str:
    first_string = f"{evaluation_result[OUTPUT_NAME_KEY_IDENTIFIER]}: " if OUTPUT_NAME_KEY_IDENTIFIER in evaluation_result.keys() else ""

    return (
        f"{first_string}"
        f"{evaluation_result[METRIC_KEY_IDENTIFIER].upper()} "
        f"{evaluation_result[CONDITION_KEY_IDENTIFIER].convert_to_plain_english()} "
        f"{evaluation_result[THRESHOLD_KEY_IDENTIFIER]}"
        f"{_format_plain_english_dimension_value(evaluation_result)}"
    )


def make_plain_english_inner_result(inner_evaluation_result: dict) -> str:
    metric_value = inner_evaluation_result[METRIC_VALUE_KEY_IDENTIFIER]

    return (
        f"{inner_evaluation_result[DIMENSION_VALUE_KEY_IDENTIFIER]}"
        f"{' (' + str((round(metric_value, 4))) + ')' if metric_value is not None else ': '}"
        f"{str(inner_evaluation_result[MESSAGE_KEY_IDENTIFIER]) if MESSAGE_KEY_IDENTIFIER in inner_evaluation_result.keys() else ''}"
    )


def _format_plain_english_dimension_value(evaluation_result: dict) -> str:
    if "dimension_value" in evaluation_result.keys():
        dim_values = evaluation_result[DIMENSION_VALUE_KEY_IDENTIFIER]
        dim_values = _make_str_or_list_of_str(dim_values)
        if isinstance(dim_values, list):
            return f" for {enumerate_elements_in_a_sentence(dim_values)}"
        else:
            return f" for {dim_values}"

    elif "dimension" in evaluation_result.keys():
        return f" per {evaluation_result[DIMENSION_KEY_IDENTIFIER]}"

    return ""


def _make_str_or_list_of_str(thing):
    if isinstance(thing, str):
        return thing

    elif isinstance(thing, list):
        return [_make_str_or_list_of_str(i) for i in thing]

    return str(thing)



from datetime import datetime
from typing import Dict, List, Union

import holidays
import numpy as np
import pandas as pd

from inspec_ai.preprocessing.predictive_features import get_predictive_features

DATES_NOT_A_PD_SERIES_ERROR = "The dates input should be a pandas series, not '{input_type}'."

DATES_NOT_A_DATETIME_DTYPE_ERROR = "The dates input have a datetime dtype, not '{input_dtype}'."

INVALID_PERIODICITY_ERROR = "The periodicity input should be either 'D', 'W' or 'M', not '{periodicity}'."

CUSTOM_DATES_NOT_DICT_ERROR = "The custom_dates input should be a dictionary, not '{input_type}'."

CUSTOM_DATES_VALUE_INVALID_TYPE_ERROR = (
    "The custom_dates input dict values should be either a list, a string with the 'YYYY-MM--DD' format or a datetime object, "
    "not '{custom_dates_value_type}'."
)

CUSTOM_DATES_LIST_VALUE_WITH_INVALID_ELEMENTS_ERROR = (
    "The elements of a list custom_dates input dict values should be either a string with the 'YYYY-MM--DD' format or a datetime object,"
    "not '{custom_dates_list_value_type}'."
)


def get_predictive_time_features(
    dates: pd.Series,
    y: Union[pd.Series, np.ndarray],
    periodicity="D",
    custom_dates: Dict[str, Union[List[Union[str, datetime]], str, datetime]] = None,
    alpha=0.1,
) -> pd.DataFrame:
    """Generates and selects the relevant time features for predicting the y series.

    The time features are automatically generated using the inspec_ai.preprocessing.time_enhancement function.
    The relevant features are then selected with a lasso regression model.

    Args:
        dates: Series of dates, should be a pandas Series with a datetime dtype.
        y: Numpy array or pandas Series. Predicted target values.
        periodicity: Periodicity of the series, should be either 'D' (daily), 'W' (weekly), or 'M' (monthly).
        custom_dates: A dictionary containing user defined important dates to add as one-hot encoded variables. Each key
            should be the name of an event and the value its date. The valid formats for the values are: list, datetime
            and str following the 'YYYY-MM-DD' format. The valid formats for the elements of a list value are: datetime and
            str with the 'YYYY-MM--DD' format.
        alpha: The L2 regularization parameter for the Lasso model. If 0, then all the features will be used.

    Returns: A pandas DataFrame containing the predictive time and seasonal features.

    """
    time_features = time_enhancement(dates, periodicity, custom_dates)
    _, predictive_time_features_coefs = get_predictive_features(time_features, y, alpha)

    predictive_columns = [col for col, coef in predictive_time_features_coefs.items() if abs(coef) > 0]

    return time_features[predictive_columns]


def time_enhancement(
    dates: pd.Series,
    periodicity="D",
    custom_dates: Dict[str, Union[List[Union[str, datetime]], str, datetime]] = None,
) -> pd.DataFrame:
    """Generates a dataframe with a set of one-hot encoded variables accounting for time and seasonal events based on a series of dates.

    Args:
        dates: Series of dates, should be a pandas Series with a datetime dtype.
        periodicity: Periodicity of the series, should be either 'D' (daily), 'W' (weekly), or 'M' (monthly).
        custom_dates: A dictionary containing user defined important dates to add as one-hot encoded variables. Each key
            should be the name of an event and the value its date. The valid formats for the values are: list, datetime
            and str following the 'YYYY-MM-DD' format. The valid formats for the elements of a list value are: datetime and
            str with the 'YYYY-MM--DD' format.

    Returns: A pandas DataFrame containing the time and seasonal features.
    """
    if not custom_dates:
        custom_dates = {}

    _validate_dates(dates)
    _validate_custom_dates(custom_dates)

    if periodicity == "D":
        return _time_enhancement_daily(dates, custom_dates)

    elif periodicity == "W":
        return _time_enhancement_weekly(dates)

    elif periodicity == "M":
        return _time_enhancement_monthly(dates)

    else:
        raise ValueError(INVALID_PERIODICITY_ERROR.format(periodicity=periodicity))


def _time_enhancement_daily(dates: pd.Series, custom_dates: Dict[str, Union[List[Union[str, datetime]], str, datetime]]) -> pd.DataFrame:
    """Time enhancement for daily periodicity."""

    time_features = [
        _make_week_dummies(dates),
        _make_month_dummies(dates),
        _make_day_of_week_dummies(dates),
        _make_holiday_dummies(dates, prov="QC"),
        _make_trigonometric_features(dates),
    ]

    if custom_dates:
        time_features.append(_make_custom_dates_dummies(dates, custom_dates))

    return pd.concat(time_features, axis=1)


def _time_enhancement_weekly(dates: pd.Series) -> pd.DataFrame:
    """Time enhancement for weekly periodicity."""
    time_features = [
        _make_week_dummies(dates),
        _make_month_dummies(dates),
        _make_trigonometric_features(dates),
    ]

    return pd.concat(time_features, axis=1)


def _time_enhancement_monthly(dates: pd.Series) -> pd.DataFrame:
    """Time enhancement for monthly periodicity."""
    time_features = [_make_month_dummies(dates), _make_trigonometric_features(dates)]

    return pd.concat(time_features, axis=1)


def _make_month_dummies(dates: pd.Series) -> pd.DataFrame:
    return pd.get_dummies(dates.dt.month_name())


def _make_week_dummies(dates: pd.Series) -> pd.DataFrame:
    return pd.get_dummies(dates.dt.week, prefix="Week")


def _make_day_of_week_dummies(dates: pd.Series) -> pd.DataFrame:
    return pd.get_dummies(dates.dt.day_name())


def _make_holiday_dummies(dates: pd.Series, prov="QC") -> pd.DataFrame:
    holiday_series = dates.apply(_get_canadian_holiday, prov=prov)
    holiday_dummies_df = pd.get_dummies(holiday_series, prefix="holiday")

    return holiday_dummies_df.drop(columns="holiday_None", errors="ignore")


def _make_trigonometric_features(dates: pd.Series) -> pd.DataFrame:
    share_of_year_completed = dates.dt.day_of_year / (365 + dates.dt.is_leap_year)
    angle = (share_of_year_completed * 2 * np.pi).values
    return pd.DataFrame({"sine": np.sin(angle), "cosine": np.cos(angle)})


def _make_custom_dates_dummies(
    dates: pd.Series,
    custom_dates: Dict[str, Union[List[Union[str, datetime]], str, datetime]],
) -> pd.DataFrame:
    dummies = {}

    for event_name, day_or_list_of_days in custom_dates.items():

        if isinstance(day_or_list_of_days, list):
            days = day_or_list_of_days
            datetime_dates = [_to_datetime(day).date() for day in days]
            dummies[event_name] = dates.dt.date.isin(datetime_dates).astype(int)

        else:
            day = day_or_list_of_days
            dummies[event_name] = (dates.dt.date == _to_datetime(day).date()).astype(int)

    return pd.DataFrame(dummies)


def _to_datetime(date: Union[str, datetime]) -> datetime:
    if isinstance(date, str):
        return datetime.strptime(date, "%Y-%m-%d")

    return date


def _get_canadian_holiday(date: str, prov="QC") -> str:
    holidays_dict_like = holidays.Canada(prov=prov)

    try:
        return holidays_dict_like[date].replace(" (Observed)", "")

    except KeyError:
        return "None"


def _validate_dates(dates: pd.Series):
    if not isinstance(dates, pd.Series):
        raise ValueError(DATES_NOT_A_PD_SERIES_ERROR.format(input_type=type(dates)))

    if "datetime" not in str(dates.dtype):
        raise ValueError(DATES_NOT_A_DATETIME_DTYPE_ERROR.format(input_dtype=dates.dtype))


def _validate_custom_dates(custom_dates: Dict[str, Union[List[Union[str, datetime]], str, datetime]]):
    if isinstance(custom_dates, dict):

        for custom_dates_value in custom_dates.values():

            if isinstance(custom_dates_value, list):
                for date in custom_dates_value:
                    if not isinstance(date, str) and not isinstance(date, datetime):
                        raise ValueError(CUSTOM_DATES_LIST_VALUE_WITH_INVALID_ELEMENTS_ERROR.format(custom_dates_list_value_type=type(date)))

            else:
                if not isinstance(custom_dates_value, str) and not isinstance(custom_dates_value, datetime):
                    raise ValueError(
                        CUSTOM_DATES_VALUE_INVALID_TYPE_ERROR.format(
                            custom_dates_value_type=type(custom_dates_value),
                        )
                    )

    else:
        raise ValueError(CUSTOM_DATES_NOT_DICT_ERROR.format(input_type=type(custom_dates)))



from inspec_ai.quality.evaluate_quality_standards import evaluate_quality_standards
from inspec_ai.quality.quality_standard import QualityStandard

__all__ = [evaluate_quality_standards, QualityStandard]



import datetime
import logging
import re
from typing import Union

import pandas as pd
import requests
from requests.adapters import HTTPAdapter, Retry


class Weather:
    def __init__(
        self,
        access_token: str,
    ):
        self.headers = {"Authorization": f"Bearer {access_token}", "Content-Type": "application/json", "accept": "application/json"}

        self.url = "https://inspec-meteo-dev-app.azurewebsites.net/Weather"

        self.periodicities = ["H", "D", "W", "M", "W-MON", "W-TUE", "W-WED", "W-THU", "W-FRI", "W-SAT", "W-SUN"]

        self.possible_features = [
            "temperature",
            "wind",
            "temperaturefeels",
            "dewpoint",
            "humidity",
            "pressure",
            "winddir",
            "windspeedgust",
            "sky",
            "precipitation",
            "snow",
            "solarrad",
            "weathercoded",
            "weatherprimarycoded",
            "cloudscoded",
            "isday",
        ]

        self.periodicity_features = {
            "H": [
                "temperature",
                "wind",
                "temperaturefeels",
                "dewpoint",
                "humidity",
                "pressure",
                "winddir",
                "windspeedgust",
                "sky",
                "precipitation",
                "snow",
                "solarrad",
                "weathercoded",
                "weatherprimarycoded",
                "cloudscoded",
                "isday",
            ],
            "D": [
                "temperature",
                "wind",
                "temperaturefeels",
                "dewpoint",
                "humidity",
                "pressure",
                "winddir",
                "windspeedgust",
                "sky",
                "precipitation",
                "snow",
                "solarrad",
            ],
            "W": [
                "temperature",
                "wind",
                "temperaturefeels",
                "dewpoint",
                "humidity",
                "pressure",
                "winddir",
                "windspeedgust",
                "sky",
                "precipitation",
                "snow",
                "solarrad",
            ],
            "M": [
                "temperature",
                "wind",
                "temperaturefeels",
                "dewpoint",
                "humidity",
                "pressure",
                "winddir",
                "windspeedgust",
                "sky",
                "precipitation",
                "snow",
                "solarrad",
            ],
        }

        self.aggregations = {
            "temperature": ["avg", "max", "min", "spread"],
            "wind": ["avg", "max", "min", "spread"],
            "temperaturefeels": ["avg", "max", "min", "spread"],
            "dewpoint": ["avg", "max", "min", "spread"],
            "humidity": ["avg", "max", "min", "spread"],
            "pressure": ["avg", "max", "min", "spread"],
            "winddir": ["avg", "max", "min", "spread"],
            "windspeedgust": ["avg", "max", "min", "spread"],
            "sky": ["avg", "max", "min", "spread"],
            "precipitation": ["avg", "max", "min", "spread", "sum"],
            "snow": ["avg", "max", "min", "spread", "sum"],
            "solarrad": ["avg", "max", "min", "spread", "sum"],
        }

        self.earliest_date = datetime.datetime.fromisoformat("2011-04-01T00:00:00+00:00")

    def get_weather(
        self,
        coordinates: Union[dict, list],
        start_date: str = None,
        end_date: str = None,
        timezone_offset=None,
        periodicity="H",
        features=None,
        verbose=True,
    ):

        """Retrieve weather information
           Args:
               coordinates: list of dict or one dictionary with lat: double representing the latitude of the desired location
                            and long: double representing the longitude of the desired location, {'lat':45,'long':-75}
               start_date: string in iso format: YYYY-MM-DDTHH:MM:SS-HH:MM or YYYY-MM-DDTHH:MM:SS+HH:MM
               end_date: start_date: string in iso format: YYYY-MM-DDTHH:MM:SS-HH:MM or YYYY-MM-DDTHH:MM:SS+HH:MM
               timezone_offset: string for the timezone offset, ex: +01:30 or -05:00
               periodicity: string representing the time aggregation. Choice is among: "H", "D", "W", "M",
               "W-MON", "W-TUE", "W-WED", "W-THU", "W-FRI", "W-SAT", "W-SUN"
               features: list of features
               verbose: Whether to print the detailed results when available.
        Returns: A dataframe containing the results aggregated
        """

        start_date, end_date, coordinates, timezone_offset, features = self._normalize_parameters(
            start_date, end_date, coordinates, timezone_offset, features, periodicity
        )

        start_str, end_str, coordinates, periodicity, features, timezone_offset = self._verify_query(
            start_date, end_date, coordinates, periodicity, features, timezone_offset
        )

        query = []
        offset = timezone_offset[0]
        for i in range(len(coordinates)):
            el = coordinates[i]
            if len(timezone_offset) > 1:
                offset = timezone_offset[i]
            query.append(
                {
                    "from": f"{start_str}",
                    "to": f"{end_str}",
                    "lat": el["lat"],
                    "long": el["long"],
                    "periodicity": periodicity,
                    "timeOffset": offset,
                    "features": features,
                }
            )

        logging.basicConfig(level=logging.DEBUG)
        s = requests.Session()
        retries = Retry(total=5, backoff_factor=1, status_forcelist=[502, 503, 504], method_whitelist=["POST"])
        s.mount("http://", HTTPAdapter(max_retries=retries))

        response = s.post(self.url, headers=self.headers, json=query)

        if response.status_code in [401, 403]:
            raise Exception("Please verify that the token used when constructing the Weather class is valid")

        try:
            dfs = [pd.json_normalize(data) for data in response.json()]
            df = pd.concat(dfs)
            df.columns = [col.split(".")[1] if len(col.split(".")) > 1 else col.split(".")[0] for col in df.columns]

            df["datetime"] = pd.to_datetime(df["datetime"])
            df["is_forecast"] = df["datetime"] > datetime.datetime.utcnow().replace(tzinfo=datetime.timezone.utc)

            if verbose:
                print(f"Response time: {response.elapsed.total_seconds()}")
                print(f"Successfully fetched {df.shape[0]} values for {df.shape[1] - 4} features")
                if len(df.loc[:, df.isna().any()].columns):
                    print("The columns", df.loc[:, df.isna().any()].columns.to_list(), "contain null values")
                    print("The number of rows containing null values is: ", df[df.isna().any(axis=1)].shape[0])
                    print("The dates which contain null values are: ", df[df.isna().any(axis=1)]["datetime"].to_list())
                else:
                    print("The dataframe contains no null values")
            return df

        except Exception:
            raise Exception(response.text)

    # Normalize parameters
    def _normalize_parameters(self, start_date, end_date, coordinates, timezone_offset, features, periodicity):

        if coordinates is None:
            raise Exception("Please provide coordinates")

        if type(coordinates) is dict:
            coordinates = [coordinates]

        utc_now = datetime.datetime.utcnow()

        if start_date is None:
            start_date = (datetime.datetime.utcnow() - datetime.timedelta(7)).isoformat()
        if end_date is None:
            end_date = utc_now.isoformat()

        if timezone_offset is None:
            timezone_offset = "+00:00"
            print(
                "\nWARNING: No timezone offset was specified, default set to UTC+0. The daily, weekly and monthly "
                "aggregations will be made using the start and end of the day in UTC+0 time."
            )

        if type(timezone_offset) is str:
            timezone_offset = [timezone_offset]

        if periodicity.upper() not in self.periodicities:
            raise Exception(f"Periodicity {periodicity} not supported")
        if features is None:
            features = self.periodicity_features[periodicity.split("-")[0].upper()]

        return start_date, end_date, coordinates, timezone_offset, features

    # Verify request parameters
    def _verify_query(self, start_date, end_date, coordinates, periodicity, features, timezone_offset):

        start_str, end_str = self._check_start_end_dates(start_date, end_date)

        for tz_offset in timezone_offset:
            try:
                sign, hours, minutes = re.match(r"([+\-]?)(\d{2}):(\d{2})", tz_offset).groups()
                sign = -1 if sign == "-" else 1
                hours, minutes = int(hours), int(minutes)
                datetime.timezone(sign * datetime.timedelta(hours=hours, minutes=minutes))
            except Exception:
                raise Exception(f"The timezone offset: {tz_offset} was not provided in the correct format, eg: +03:00")

        self._check_coordinates(coordinates)

        if len(timezone_offset) not in [1, len(coordinates)]:
            raise Exception(
                "Please provide either one timezone offset that will be applied to all locations "
                "or a list of timezone offsets corresponding to each coordinates, in the same order."
            )

        period_parts = periodicity.split("-")

        if features is not None:
            for feat in features:
                parts = feat.split("-")
                if parts[0].lower() not in self.possible_features:
                    raise Exception(f"Feature {parts[0]} not available")
                if parts[0].lower() not in self.periodicity_features[period_parts[0].upper()]:
                    raise Exception(f"Feature {parts[0]} not available for this periodicity")
                if len(parts) > 1:
                    if parts[1].lower() not in self.aggregations[parts[0].lower()]:
                        raise Exception(f"Aggregation {parts[1]} not available for {parts[0]}")

        return start_str, end_str, coordinates, periodicity, features, timezone_offset

    def _check_coordinates(self, coordinates):

        if type(coordinates) is not list:
            raise Exception("Coordinates must be a dictionary containing the latitude and longitude or a list of coordinates")

        if len(coordinates) == 0:
            raise Exception("Please provide coordinates")

        for element in coordinates:
            if (len(element.keys()) != 2) or ("lat" not in element.keys()) or ("long" not in element.keys()):
                raise Exception(f"The provided location {element} is invalid, eg of correct format: {{'lat': 43,'long': 18}}")
            if element["lat"] > 90 or element["lat"] < -90 or element["long"] > 180 or element["long"] < -180:
                raise Exception(f"The provided location is invalid: lat: {element['lat']}, long: {element['long']}")

    def _check_start_end_dates(self, start_date, end_date):
        try:
            start_obj, end_obj = datetime.datetime.fromisoformat(start_date), datetime.datetime.fromisoformat(end_date)
        except ValueError:
            raise Exception("Please provide the start and end date in isoformat, ex:2022-05-25T14:25:00 ")

        utc_now = datetime.datetime.utcnow()

        max_forecast = (utc_now + datetime.timedelta(days=15)).replace(tzinfo=datetime.timezone.utc)
        end_obj_utc = (datetime.datetime.utcfromtimestamp(end_obj.timestamp())).replace(tzinfo=datetime.timezone.utc)

        LOCAL_TIMEZONE = datetime.datetime.now().astimezone().tzinfo
        if not start_obj.tzinfo:
            start_obj = start_obj.replace(tzinfo=LOCAL_TIMEZONE)
            print(f"\nWARNING: No timezone was specified for the start date, default set to your local time zone: {start_obj.strftime('%Z %z')}")

        # Save timezone info
        end_tzinfo = end_obj.tzinfo
        if not end_tzinfo:
            end_obj = end_obj.replace(tzinfo=LOCAL_TIMEZONE)
            print(f"\nWARNING: No timezone was specified for the end date, default set to your local time zone: {end_obj.strftime('%Z %z')}")

        if end_obj_utc > max_forecast:
            end_obj = max_forecast
            if not end_tzinfo:
                end_obj = end_obj.astimezone(LOCAL_TIMEZONE)
            else:
                end_obj = end_obj.astimezone(end_tzinfo)
            print(f"\nWARNING: Forecast only available until {end_obj}")

        if start_obj < self.earliest_date:
            raise Exception(f"Historical data is not available before {self.earliest_date}")

        if start_obj > end_obj:
            raise Exception("Start date must be earlier than end date")

        return start_obj.isoformat(), end_obj.isoformat()



import json
import os
import pathlib
import sys
import threading
import time
import uuid
from datetime import datetime, timedelta
from queue import Queue
from typing import Dict

import requests

from inspec_ai.config import TRACKING_DISABLED

ANONYMOUS_TRACKING_IDENTIFIER_KEY = "anonymous_tracking_identifier"

telemetry_queue: Queue = Queue()

previously_tracked_events: Dict = {}

_anonymous_tracking_identifier = ""


def track_function_executed(function_name, *args, **kwargs):
    def decorate(func):
        def call(*args, **kwargs):
            _track_function_executed_internal(function_name)
            result = func(*args, **kwargs)
            return result

        return call

    return decorate


def _track_function_executed_internal(function_name: str) -> None:
    if TRACKING_DISABLED:
        return

    utc_now = datetime.utcnow()

    if function_name in previously_tracked_events:
        last_track = previously_tracked_events[function_name]

        previously_tracked_events[function_name] = utc_now

        if utc_now < last_track + timedelta(minutes=5):
            return

    else:
        previously_tracked_events[function_name] = utc_now

    telemetryEvent = {
        "eventDate": str(utc_now),
        "functionName": str(function_name),
    }

    telemetry_queue.put(telemetryEvent)


def _get_anonymous_tracking_identifier() -> str:
    global _anonymous_tracking_identifier

    if _anonymous_tracking_identifier and _anonymous_tracking_identifier != "":
        return _anonymous_tracking_identifier

    data_dir = _get_datadir()
    directory = os.path.join(data_dir, ".inspec")
    file_name = "config.json"

    full_path = os.path.join(directory, file_name)

    if os.path.isfile(full_path):
        with open(full_path, "r") as file:
            config = json.load(file)

    else:
        if not os.path.exists(directory):
            os.mkdir(directory)

        anonymous_tracking_identifier = uuid.uuid4()

        config = {ANONYMOUS_TRACKING_IDENTIFIER_KEY: str(anonymous_tracking_identifier)}

        with open(full_path, "w") as file:
            json.dump(config, file)

    _anonymous_tracking_identifier = config[ANONYMOUS_TRACKING_IDENTIFIER_KEY]

    return _anonymous_tracking_identifier


def _get_datadir() -> pathlib.Path:
    home = pathlib.Path.home()

    if sys.platform == "win32":
        return home / "AppData/Roaming"
    elif sys.platform == "linux":
        return home / ".local/share"
    elif sys.platform == "darwin":
        return home / "Library/Application Support"


def _telemetry_daemon():
    main_process_alive = True
    while main_process_alive:
        mt = threading.main_thread()
        main_process_alive = mt.is_alive()

        while not telemetry_queue.empty():
            event = telemetry_queue.get()

            anonymous_tracking_identifier = _get_anonymous_tracking_identifier()

            event["anonymousTelemetryIdentifier"] = str(anonymous_tracking_identifier)

            try:
                requests.post(
                    "https://inspec-dev-api.azurewebsites.net/telemetry/track-function-executed",
                    json=event,
                    headers={"Authorization": "Bearer NDk5MDYxRjctQjM4Mi00MkRBLTlFNUMtNjdCOTBBMTBDM0VD"},
                )
            except:  # noqa: E722 - we want to swallow the error and not act on it.
                pass

        time.sleep(0.1)


telemetry_thread = threading.Thread(target=_telemetry_daemon, name="Telemetry Thread")
telemetry_thread.start()



import itertools
import warnings
from copy import deepcopy
from typing import Dict, List, Tuple, Union

import numpy as np
import pandas as pd
from sklearn.impute import SimpleImputer
from sklearn.linear_model import LinearRegression
from sklearn.metrics import mean_squared_error

from inspec_ai.analytics import telemetry_service

KNOWN_IMPUTATION_STRATEGIES = ["mean", "median", "most_frequent"]
X_AND_Y_INVALID_FORMAT_ERROR = "The input datasets x and y should be either numpy arrays or pandas dataframes."


@telemetry_service.track_function_executed("fill_missing_values")
def fill_missing_values(
    x: Union[pd.DataFrame, pd.Series, np.ndarray],
    y: Union[pd.DataFrame, pd.Series, np.ndarray],
    replacement_strategies: Dict = None,
) -> Tuple[Union[pd.DataFrame, pd.Series, np.ndarray], Dict]:
    """Fills the dataset's missing values using the optimal strategy for each column.

    Requires the x and y datasets to be only filled with numerical values and nans.
    Assuming that the optimal strategy for a simple model is also optimal for a more complex model,
    the function uses a grid search strategy with a linear regression model.

    Args:
        x: Training data with numerical and np.nan values
        y: Target numerical values.
        replacement_strategies: Optional. A dictionary allowing the user to select the missing values replacement strategy.
            The keys should refer to x column names and values to corresponding replacement strategies.
            The available strategies are: 'mean', 'median' and 'most_frequent'.
            The function will determine the optimal strategies for the remaining columns.

    Returns: The dataset with filled missing values and a dictionary containing containing the applied strategy for each column with missing values.

    """
    if isinstance(x, pd.DataFrame):
        return _fill_missing_values_pandas_dataframe(x, y, replacement_strategies)

    elif isinstance(x, pd.Series):
        return _fill_missing_values_pandas_series(x, y, replacement_strategies)

    elif isinstance(x, np.ndarray):
        return _fill_missing_values_numpy(x, y, replacement_strategies)

    else:
        raise ValueError(X_AND_Y_INVALID_FORMAT_ERROR)


def _fill_missing_values_pandas_dataframe(x: pd.DataFrame, y: pd.DataFrame, replacement_strategies: Dict = None) -> Tuple[pd.DataFrame, Dict]:
    # Step 0: Setup
    x_tmp = deepcopy(x)
    y_tmp = deepcopy(y)

    if not replacement_strategies:
        replacement_strategies = {}

    _validate_replacement_strategies(list(x_tmp.columns), replacement_strategies)

    selected_strategies = dict((col, None) for col in x_tmp.columns)
    column_has_missing_values = np.isnan(x_tmp).any()

    if (column_has_missing_values == False).all():
        warnings.warn("The dataset has no missing values, returned as is.")
        return x, {}

    # Step 1: Fill missing values with the user's selected strategies
    x_tmp, applied_strategies = _apply_imputation_strategies(x_tmp, replacement_strategies)
    selected_strategies.update(applied_strategies)

    # Step 2: Check if there are still any admissible columns
    (
        admissible_columns,
        columns_with_all_missing_values,
    ) = _get_admissible_columns_for_imputation_and_columns_with_all_missing_values(x_tmp)

    if not admissible_columns:
        return x_tmp, replacement_strategies

    # Step 3: Automatically impute columns
    x_without_all_missing_values_columns = x_tmp.drop(columns=columns_with_all_missing_values, inplace=False)
    (
        imputed_variables,
        best_strategies_combination,
    ) = _find_and_apply_the_best_imputation_strategies(x_without_all_missing_values_columns, y_tmp, admissible_columns)

    for col in admissible_columns:
        x_tmp[col] = imputed_variables[col]

    # Step 4: format output
    formatted_best_strategies = {admissible_columns[i]: best_strategies_combination[i] for i in range(len(best_strategies_combination))}
    selected_strategies.update(formatted_best_strategies)

    return x_tmp, selected_strategies


def _fill_missing_values_pandas_series(x: pd.Series, y: pd.Series, replacement_strategies: Dict = None) -> Tuple[pd.Series, Dict]:

    if not x.name:
        x.name = 0

    x_tmp, selected_strategies = _fill_missing_values_pandas_dataframe(pd.DataFrame(x), pd.DataFrame(y), replacement_strategies)

    return x_tmp.iloc[:, 0], selected_strategies


def _fill_missing_values_numpy(x: np.ndarray, y: np.ndarray, replacement_strategies: Dict[int, str] = None) -> Tuple[np.ndarray, Dict]:
    x_tmp, selected_strategies = _fill_missing_values_pandas_dataframe(pd.DataFrame(x), pd.DataFrame(y), replacement_strategies)
    return x_tmp.values, selected_strategies


def _impute_values_of_series(values, strategy):
    column_values = deepcopy(values)
    column_values = column_values.values.reshape(-1, 1)

    imputer = SimpleImputer(strategy=strategy)
    imputer.fit(column_values)

    return imputer.transform(column_values)


def _apply_imputation_strategies(x: pd.DataFrame, replacement_strategies: Dict):
    column_has_missing_values = np.isnan(x).any()
    column_is_all_missing_values = np.isnan(x).all()
    applied_strategies = {}

    for col_name, strategy in replacement_strategies.items():
        if column_is_all_missing_values[col_name]:
            warnings.warn(f"Could not fill missing values for column '{col_name}' with strategy '{strategy}' since it does not contain any non-missing value.")

        elif not column_has_missing_values[col_name]:
            warnings.warn(f"Could not fill missing values for column '{col_name}' with strategy '{strategy}' since it does have any missing value.")

        else:
            x[col_name] = _impute_values_of_series(x[col_name], strategy)
            applied_strategies[col_name] = strategy

    return x, applied_strategies


def _get_admissible_columns_for_imputation_and_columns_with_all_missing_values(
    x: pd.DataFrame,
):
    column_has_missing_values = np.isnan(x).any()
    column_is_all_missing_values = np.isnan(x).all()

    column_is_admissible_for_imputation = column_has_missing_values & -column_is_all_missing_values

    admissible_columns = list(column_is_admissible_for_imputation[column_is_admissible_for_imputation].index)

    columns_with_all_missing_values = list(column_is_all_missing_values[column_is_all_missing_values].index)

    for column in column_is_all_missing_values[column_is_all_missing_values].index:
        warnings.warn(f"Could not fill missing values for column '{column}' since it does not contain any non-missing value.")

    return admissible_columns, columns_with_all_missing_values


def _find_and_apply_the_best_imputation_strategies(x: pd.DataFrame, y: pd.DataFrame, columns: List[str]):
    known_imputations: Dict = {col: {} for col in columns}

    for column in columns:
        for strategy in KNOWN_IMPUTATION_STRATEGIES:
            known_imputations[column][strategy] = _impute_values_of_series(x[column], strategy)

    # Find best imputation strategies combination (grid search)
    strategies_combinations = _lists_cartesian_product([KNOWN_IMPUTATION_STRATEGIES] * len(columns))

    rmse = {}

    for combination in strategies_combinations:
        trial_x = deepcopy(x)

        for index, strategy in enumerate(combination):
            col = columns[index]
            trial_x[col] = known_imputations[col][strategy]

        trial_model = LinearRegression()
        trial_model.fit(trial_x, y)
        trial_rmse = mean_squared_error(y, trial_model.predict(trial_x))

        rmse[trial_rmse] = combination

    min_rmse = min(list(rmse.keys()))
    best_strategies_combination = rmse[min_rmse]

    for index, strategy in enumerate(best_strategies_combination):
        col = columns[index]
        x[col] = known_imputations[col][strategy]

    return x[columns], best_strategies_combination


def _lists_cartesian_product(list_of_lists):
    out = []

    for element in itertools.product(*list_of_lists):
        out.append(element)

    return out


def _validate_replacement_strategies(x_columns: List, replacement_strategies: Dict):
    for col_name in replacement_strategies.keys():
        if col_name not in x_columns:
            raise ValueError(
                f"'{col_name}' is not a column of the x dataset. " f"Please ensure that the keys of replacement_strategies are columns of the x dataset."
            )

    for strategy in replacement_strategies.values():
        if strategy not in KNOWN_IMPUTATION_STRATEGIES:
            raise ValueError(
                f"'{strategy}' is not an available missing value replacement strategy. Please ensure that"
                f"the values of replacement_strategies are valid replacement strategies. The valid"
                f"replacement strategies are: {str(KNOWN_IMPUTATION_STRATEGIES)[1:-1]}"
            )



from typing import Optional, Type, Union

from inspec_ai.quality.enums import Condition, EvaluationMode, Metric
from inspec_ai.quality.exceptions import ConfigError
from inspec_ai.utils.text_utils import enumerate_elements_in_a_sentence
from inspec_ai.utils.utils import _StrEnum


class QualityStandard:
    """Quality standard for the evaluate_quality_standards function.

    Attributes:
        metric: The error measure to evaluate the model on. Can be 'mae', 'mase', 'mape', 'auc', 'f1', 'precision', 'recall', or 'accuracy'.
            The init method accepts a string, but sets the attribute as an enum.
        threshold: The limit (numerical) value that the metric should be.
            Whether the metric should be equal, lower or higher than the threshold is defined by the condition attribute.
        condition: An indicator whether the error should be equal ('eq'), greater ('ht'), greater or equal ('gte'), lower ('lt'),
            or lower or equal ('lte') to the threshold. The init method accepts a string, but sets the attribute as an enum.
        dimension: Indicates a dimension column for which the error measure should be evaluated. If the dimension_value attribute is None,
            then the evaluation mode will be set to 'for each unique dimension value', i.e. the quality standard should be evaluated for each group.
            If the dimension_value is defined, then the quality standard will only be evaluated for that dimension value, or that group of dimension values.
            If the dimension is set to None, then the quality standard should be evaluated for the whole dataset.
        dimension_value: Indicates a dimension value or a group or dimension values for which the quality standard should be evaluated.
            he dimension_value should only be set if the dimension attribute is also set (not None).
        name: Optional quality standard name.
        evaluation_mode: Indicates whether the quality standard should be evaluated for the whole dataset ("global"), for each dimension in the dataset,
            for one dimension, or for each dimension in a list
    """

    def __init__(
        self,
        metric: str,
        threshold: Union[int, float],
        condition: str,
        dimension: Optional[Union[str, int, float, bool]] = None,
        dimension_value: Optional[Union[list, str, int, float, bool]] = None,
        output_name: Optional[str] = None,
        name: Optional[str] = None,
    ):
        self._check_if_user_input_in_allowed_values(metric, Metric)
        self._check_if_user_input_in_allowed_values(condition, Condition)

        self.metric: Metric = Metric(metric)
        self.output_name = output_name
        self.threshold = threshold
        self.condition: Condition = Condition(condition)
        self.dimension = dimension
        self.dimension_value = dimension_value
        self.name = name
        self.evaluation_mode: EvaluationMode = self._determine_evaluation_mode()

    def _determine_evaluation_mode(self):
        accepted_types_single_values = (int, float, str, bool)

        if self.dimension:
            if self.dimension_value:
                if isinstance(self.dimension_value, accepted_types_single_values):
                    return EvaluationMode.FOR_SINGLE_DIM_VALUE
                else:
                    return EvaluationMode.FOR_EACH_DIM_VALUE_IN_LIST
            else:
                return EvaluationMode.FOR_EACH_UNIQUE_DIM_VALUE
        return EvaluationMode.GLOBAL

    @staticmethod
    def _check_if_user_input_in_allowed_values(user_input: str, enum: Type[_StrEnum]):
        allowed_values = [str(value) for value in enum]

        if user_input not in allowed_values:
            raise ConfigError(
                f"Unknown '{enum.__name__.lower()}': '{user_input}'. "
                f"Please specify one of the following metrics: {enumerate_elements_in_a_sentence(allowed_values).replace('and', 'or')}"
            )



import glob
import pathlib
from setuptools import find_packages, setup
from inspec_ai import __version__

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()

# data files (non-python modules) to be included in the package
assets = glob.glob("inspec_ai/quality/visualisation/assets/**/*.*", recursive=True)
templates = glob.glob(
    "inspec_ai/quality/visualisation/templates/**/*.*", recursive=True
)
data_files = [("inspec_ai", assets + templates)]

setup(
    name="inspec-ai",
    version=__version__,
    license="Internal Use",  # Chose a license from here: https://help.github.com/articles/licensing-a-repository
    description="Library containing all the prototypes that were developped by the Moov AI product team.",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://moov.ai/en/",
    author="MoovAI Technologies Inc",
    author_email="info@moov.ai",
    classifiers=[
        "Development Status :: 3 - Alpha",  # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
    packages=find_packages(exclude=["tests.*", "tests"]),
    include_package_data=True,
    install_requires=[
        "colorama==0.4.4",
        "holidays==0.12",
        "jinja2==3.0.3",
        "kaggle==1.5.12",
        "matplotlib==3.5.1",
        "numpy==1.21.3",
        "pandas==1.3.4",
        "pandas-profiling==3.1.0",
        "pillow==9.1.0",
        "scikit-learn==1.0.2",
        "fuzzywuzzy==0.18.0",
        "requests==2.28.1",
    ],
    entry_points={},
    data_files=data_files,
)



import base64
import io
import os
from collections import namedtuple
from datetime import date, datetime
from pathlib import Path
from typing import List

import matplotlib.pyplot as plt
from jinja2 import Environment, FileSystemLoader
from PIL import Image

from inspec_ai.__init__ import package_root_dir
from inspec_ai.quality.constants import (
    DETAILED_RESULTS_KEY_IDENTIFIER,
    DIMENSION_KEY_IDENTIFIER,
    DIMENSION_VALUE_KEY_IDENTIFIER,
    ERROR_COUNT_KEY_IDENTIFIER,
    EVALUATIONS_RESULTS_KEY_IDENTIFIER,
    FAIL_COUNT_KEY_IDENTIFIER,
    MESSAGE_KEY_IDENTIFIER,
    METRIC_VALUE_KEY_IDENTIFIER,
    PASS_COUNT_KEY_IDENTIFIER,
    STATUS_KEY_IDENTIFIER,
)
from inspec_ai.quality.utils import make_plain_english_inner_result, make_plain_english_result

GREEN_HEX = "#00bc80"
RED_HEX = "#f93480"

VISUALISATION_DIR: Path = package_root_dir / "quality/visualisation"
TEMPLATES_DIR = VISUALISATION_DIR / "templates"
ASSETS_DIR = VISUALISATION_DIR / "assets"
ICONS_DIR = ASSETS_DIR / "icons"
FONTS_DIR = ASSETS_DIR / "fonts"

Assets = namedtuple("Assets", "green_check_img red_cross_img exclamation_point_img nunito_fonts styles")


def generate_main_html_report(evaluation_results: dict, output_folder: Path):
    assets = _load_assets()

    completion_circle_img = _make_base64_completion_circle(
        evaluation_results[PASS_COUNT_KEY_IDENTIFIER], evaluation_results[FAIL_COUNT_KEY_IDENTIFIER], evaluation_results[ERROR_COUNT_KEY_IDENTIFIER]
    )

    html_report = _render_html_report(evaluation_results, assets, completion_circle_img)

    _save_html_report(html_report, output_folder)


def _render_html_report(evaluation_results: dict, assets: Assets, completion_circle_img):
    env = Environment(loader=FileSystemLoader(TEMPLATES_DIR))
    template = env.get_template("main_report.html")

    return template.render(
        date=date.today().strftime("%d/%m/%Y"),
        pass_count=evaluation_results[PASS_COUNT_KEY_IDENTIFIER],
        n_evaluations=len(evaluation_results[EVALUATIONS_RESULTS_KEY_IDENTIFIER]),
        evaluation_results=_format_results_for_html_report(evaluation_results),
        completion_circle_img=completion_circle_img,
        green_check_img=assets.green_check_img,
        red_cross_img=assets.red_cross_img,
        exclamation_point_img=assets.exclamation_point_img,
        nunito_fonts=assets.nunito_fonts,
        styles=assets.styles,
    )


def _save_html_report(html_report: str, output_folder: Path):
    output_path = output_folder / f"quality_standards_evaluation_{datetime.today().strftime('%Y_%m_%d_%H_%M_%S')}.html"

    with open(output_path, "w") as writer:
        writer.write(html_report)


def _load_assets() -> Assets:
    green_check_img = _load_image_as_base64(ICONS_DIR / "green_check.png")
    red_cross_img = _load_image_as_base64(ICONS_DIR / "red_cross.png")
    exclamation_point_img = _load_image_as_base64(ICONS_DIR / "exclamation_point.png")
    nunito_fonts = _load_font_family(FONTS_DIR / "nunito-v23-latin")

    with open(ASSETS_DIR / "styles.css", "r") as reader:
        styles = reader.read()

    return Assets(green_check_img, red_cross_img, exclamation_point_img, nunito_fonts, styles)


def _format_results_for_html_report(evaluation_results: dict) -> List[dict]:
    results_for_html_report = []

    for result in evaluation_results[EVALUATIONS_RESULTS_KEY_IDENTIFIER]:

        result_name = result.get("name") if result.get("name") else make_plain_english_result(result)

        if (DIMENSION_KEY_IDENTIFIER not in result.keys() and DIMENSION_VALUE_KEY_IDENTIFIER not in result.keys()) or (
            DIMENSION_VALUE_KEY_IDENTIFIER in result.keys() and not isinstance(result[DIMENSION_VALUE_KEY_IDENTIFIER], list)
        ):
            result_name += f" ({round(result[METRIC_VALUE_KEY_IDENTIFIER],2)})"

        result_for_html_report = {
            "result_name": result_name,
            STATUS_KEY_IDENTIFIER: result[STATUS_KEY_IDENTIFIER].lower(),
        }

        if DETAILED_RESULTS_KEY_IDENTIFIER in result.keys():
            inner_results = []

            for inner_result in result[DETAILED_RESULTS_KEY_IDENTIFIER]:
                inner_results.append(
                    {"result_name": make_plain_english_inner_result(inner_result), STATUS_KEY_IDENTIFIER: inner_result[STATUS_KEY_IDENTIFIER].lower()}
                )

            result_for_html_report[DETAILED_RESULTS_KEY_IDENTIFIER] = inner_results

        if MESSAGE_KEY_IDENTIFIER in result.keys():
            result_for_html_report[MESSAGE_KEY_IDENTIFIER] = result[MESSAGE_KEY_IDENTIFIER]

        results_for_html_report.append(result_for_html_report)

    return results_for_html_report


def _make_base64_completion_circle(pass_count: int, fail_count: int, error_count: int) -> str:
    data = [pass_count, fail_count, error_count]
    colors = [GREEN_HEX, RED_HEX, "#ffffff"]

    pie_sections, _ = plt.pie(data, colors=colors, wedgeprops={"edgecolor": "k", "linewidth": 5})

    pie_sections[2].set_hatch("/")
    plt.rcParams["hatch.linewidth"] = 3

    if error_count > 0:
        ax = plt.gca()
        pos = ax.get_position()
        ax.set_position([pos.x0, pos.y0, pos.width * 0.9, pos.height])
        ax.legend(handles=pie_sections, labels=["pass", "fail", "error"], fontsize=24, frameon=False, bbox_to_anchor=(1, 0.75))

    buffer = io.BytesIO()

    plt.margins(0, 0)
    plt.savefig(buffer, format="png", transparent=True, bbox_inches="tight", pad_inches=0)

    buffer.seek(0)
    encoded_completion_circle = base64.b64encode(buffer.getvalue())
    buffer.close()

    return encoded_completion_circle.decode("utf-8")


def _load_image_as_base64(filepath: Path) -> str:
    image = Image.open(filepath)
    buffer = io.BytesIO()
    image.save(buffer, "PNG")

    buffer.seek(0)
    encoded_image = base64.b64encode(buffer.getvalue())
    buffer.close()

    return encoded_image.decode("utf-8")


def _load_font_family(fonts_dir: Path) -> List[str]:
    individual_fonts = []

    for file in os.listdir(fonts_dir):
        with open(fonts_dir / file, "rb") as reader:
            font = reader.read()

        utf_font = base64.b64encode(font)

        file_ext = file.split(".")[-1]

        individual_fonts.append(f"(data:application/font-{file_ext};charset=utf-8;base64,{utf_font.decode('utf8')}) format('{file_ext}')")

    return individual_fonts
